import asyncio
import logging
from typing import Any, Dict, Optional
from uuid import uuid4
import datetime

from symphonizer.interface import NodeRunnerStatus, NodeStatusCallback, NodeExecutorFunction
from symphonizer.utils import get_environment_info

logger = logging.getLogger(__name__)
verbose = False


class NodeRunner:
    """ A class that abstracts the processing af a node as a task. Loosely modelled from asyncio.Task
    and designed to run either asynchronous or synchronous executors. A node executor is any function,
    class method and including a class instance that implements the __call__ method. Only keyword
    argument parameters are supported. It is good practice to restrict the inputs and outputs of a
    node executor to those types easily serialised into and out of json.

    A NodeRunner instance is considered immutable once it has been started and cannot be reconfigured.
    As a consequence, in order to retry a task, a new instance must be created for execution, see
    NodeRunner.retry_task() for further info.

    A task executor performs the underlying processing and is any function that takes in keyword
    arguments and returns Any result type that can be serialised into json.

    async def executor_function(**kwargs: Any) -> Any:
        ...

    class Executor:
        async def __call__(self, **kwargs: Any) -> Any:
            ...

    executor = `Executor()` or `executor_function`
    result = await NodeRunner(executor, params)()
    result = await NodeRunner().prepare(params).run(executor)()

    """
    _parent_id: str | None
    _instance_id: str
    _environment_info: Dict[str, Any]
    _status: NodeRunnerStatus
    _running_task: asyncio.Task[Any] | None
    _start_time: datetime.datetime | None
    _end_time: datetime.datetime | None
    _done: bool
    _error: BaseException | None
    _result: Any

    _node_params: Dict[str, Any]
    _node_timeout: int
    _node_status_cb: NodeStatusCallback | None
    _node_executor: NodeExecutorFunction | None

    def __init__(
            self,
            node_executor: Optional[NodeExecutorFunction] = None,
            node_params: Optional[Dict[str, Any]] = None,
            node_timeout: Optional[int] = None,
            node_status_cb: Optional[NodeStatusCallback] = None,
            parent_id: Optional[str] = None,
    ):
        self._parent_id = parent_id
        self._instance_id = uuid4().hex
        self._environment_info = {}
        self._status = "new"
        self._running_task = None
        self._start_time = None
        self._end_time = None
        self._done = False
        self._error = None
        self._result = None

        self._node_executor = node_executor
        self._node_params = {} if node_params is None else node_params
        self._node_timeout = 0 if node_timeout is None else node_timeout
        self._node_status_cb = node_status_cb

        if verbose:
            logger.debug(f"New {type(self).__name__} instance -> {self._instance_id}")

    def retry_task(self) -> "NodeRunner":
        """ A NodeRunner is immutable once executed, in order to retry a task, a new instance must be created
        for execution. This method returns a new instance of the NodeRunner class, with the same parameters as
        the current instance with the _parent_id set to the current instance_id.

        If the task had been chained with prepare and or run methods, the new instance will have the same parameters
        provided by the chained methods.

        :return: self.__class__(...)
        """
        if not self._done or self._status not in ("completed", "error", "timed_out", "cancelled"):
            raise RuntimeError(
                "A retry instance of this NodeRunner is only available after the task is done"
            )
        child_task_runner = self.__class__(
            node_executor=self._node_executor,
            node_params=self._node_params,
            node_timeout=self._node_timeout,
            node_status_cb=self._node_status_cb,
            parent_id=self._instance_id,
        )
        return child_task_runner

    @property
    def instance_id(self) -> str:
        return self._instance_id

    @property
    def parent_id(self) -> str | None:
        return self._parent_id

    @property
    def done(self) -> bool:
        return self._done

    @property
    def status(self) -> NodeRunnerStatus:
        return self._status

    @status.setter
    def status(self, status: NodeRunnerStatus) -> NodeRunnerStatus:
        if self._status != status:
            if self._node_status_cb:
                self._node_status_cb(self, self._status, status)
            elif verbose:
                logger.debug(
                    f"Task status changed {self._instance_id}: {self._status} -> {status}"
                )
            self._status = status
        return self._status

    @property
    def node_info(self) -> Dict[str, Any]:
        """ Returns a dictionary of the node's task information
        """
        information = self._environment_info
        information.update({
            "instance_id": self._instance_id,
            "parent_id": self._parent_id,
            "status": self._status,
            "task_timeout": self._node_timeout,
            "task_done": self._done,
            "start_time": self._start_time,
            "end_time": self._end_time,
        })
        return information

    def cancel(self) -> None:
        if not self._done and self._running_task is not None and not self._running_task.done():
            logger.debug(f"Cancelling task {self._instance_id}")
            self._running_task.cancel(f"Request to stop task {self._instance_id}")

    def prepare(self, **params: Any) -> "NodeRunner":
        """ Prepares alternative parameters for the task executor
        A utility method to allow the task executor to be replaced with alternative parameters before the task is run.
        and before executor is called.
        This method allows for command chaining to be used. e.g.:
        - result = await AsyncTaskRunner(executor).prepare(task_params)()

        :param params:
        :return: self
        """
        if self._done or self._status not in ("new", "starting"):
            raise RuntimeError("NodeRunner has already started or is done")
        self.status = "starting"
        self._node_params = params
        return self

    def run(self, node_executor: NodeExecutorFunction) -> "NodeRunner":
        """ Provides and alternate node executor to be used when the node's task is run.
        NOTE: This method does not call the node executor, it only sets it up to be called
        when the task is run.

        This method allows for command chaining to be used. e.g.:
        - result = await AsyncTaskRunner(config).run(executor)()
        - result = await AsyncTaskRunner().prepare(params).run(executor)()

        :param node_executor:
        :return: self
        """
        if self._done or self._status not in ("new", "starting"):
            raise RuntimeError("This NodeRunner has already started or is done")
        self.status = "starting"
        self._node_executor = node_executor
        return self

    async def node_executor_wrapper(self, **params: Any) -> Any:
        """ A wrapper for calling self._node_executor. This method provides a hook for subclasses to
        override execution behavior.
        """
        try:
            if self._node_executor is None:
                raise RuntimeError("A task executor has not been set")
            return await self._node_executor(**params)
        except BaseException as err:
            if verbose:
                logger.exception(f"Task execute runtime error: {err}")
            raise err

    async def __call__(self) -> Any:
        try:
            if self._done or self._running_task is not None or self._status not in ("new", "starting"):
                raise RuntimeError("Node runner has already started or is done")
            self._start_time = datetime.datetime.now(datetime.timezone.utc)
            self._environment_info = get_environment_info()
            self._environment_info.update({
                "start_time_utc": self._start_time,
                "end_time_utc": self._end_time,
            })
            self.status = "running"
            """ NOTE: work is handled in its coroutine state
            """
            work = self.node_executor_wrapper(**self._node_params)
            if self._node_timeout:
                work = asyncio.wait_for(work, self._node_timeout)
            self._running_task = asyncio.create_task(work)
            self._result = await self._running_task
            self.status = "completed"
        except asyncio.exceptions.CancelledError as err:
            self.status = "cancelled"
            self._error = err
            # logger.debug("Nodes task cancelled, %s", self._instance_id)
        except asyncio.exceptions.TimeoutError as err:
            self.status = "timed_out"
            self._error = err
            # logger.debug(f"Node's task timed out after {self._node_timeout}s, {self._instance_id}")
        except BaseException as err:  # pylint: disable=broad-except
            self.status = "error"
            self._error = err
            # logger.exception(f"Node's task had a runtime error: {err}")
        finally:
            self._done = True
            self._end_time = datetime.datetime.now(datetime.timezone.utc)
            self._environment_info.update({
                "end_time_utc": self._end_time,
            })

        if self._error:
            raise self._error
        else:
            return self._result



import logging
from typing import Dict, Optional, Callable, Any, Literal, Type, Hashable, Set


logger = logging.getLogger(__name__)

CompositionType = Type["symphonizer.composition.Composition"]
DAGNodeType = Type["symphonizer.composition.DAGNote"]
NodeRunnerType = Type["symphonizer.node_runner.NodeRunner"]

ErrorType = Dict[str, Any]
""" ErrorType: a dictionary with the following keys:
- error: str
- description: str
"""

GraphType = Dict[Hashable, Set[Hashable]]


NodeDoneStatus = Literal["retrying", "completed", "completed_error", "error", "timed_out", "cancelled"]

NodeRunnerStatus = Literal["new", "starting", "running", "completed", "error", "timed_out", "cancelled"]
NodeStatusCallback = Callable[[NodeRunnerType, NodeRunnerStatus, NodeRunnerStatus], None]
""" NodeStatusCallback:
- instance: NodeRunner instance
- old_status
- new_status
"""
NodeExecutorFunction = Callable[..., Any]
"""
A node executor is any function that takes in keyword arguments and returns Any result type that can
be serialised into json, e.g.:
 - def node_executor_function(**kwargs) -> Any
"""


CompositionStatus = Literal["new", "starting", "running", "paused", "completed", "error", "timed_out", "cancelled"]
CompositionDoneFunction = Callable[[CompositionType, CompositionStatus, Optional[ErrorType], float], None]
""" CompositionDoneFunction: is called when the schedule is done, i.e. is not able to continue the 
processing of any more nodes. 

The function is called with the following arguments:
- instance: Composition instance
- status: CompositionStatus
- error: Optional[ErrorType]
- elapsed_time: float - in seconds
"""
NodeProcessDoneFunction = Callable[[DAGNodeType, NodeDoneStatus, Optional[BaseException]], None]
""" NodeProcessDoneFunction: is called when a node has been processed, i.e. the node has either 
completed or errored. 
NOTE: the underlying node processing / execution is immutable, once it has started and cannot be re-run. 
The NodeRunner class can return a child with NodeRunner.re_run() that can be used to process the node
again.   

The function is called with the following arguments:
- instance: DAGNote instance
- status: NodeDoneStatus
- error: Optional[BaseException]
"""


""" The Exception classes below are used to control the flow of the Composition and NodeRunner classes. 
"""


class TryAgainException(Exception):
    pass


class ContinueAfterErrorException(Exception):
    pass


class StopScheduleException(Exception):
    pass



import logging
import asyncio
import functools
from asyncio import Queue
from typing import Dict, Optional, Hashable, Any, Set, Type
import uuid
from graphlib import TopologicalSorter
import time

from symphonizer.instruments.mock import MockInstrument
from symphonizer.interface import CompositionDoneFunction, NodeProcessDoneFunction, TryAgainException, \
    ContinueAfterErrorException, StopScheduleException, ErrorType, CompositionStatus, NodeRunnerType, NodeDoneStatus, \
    GraphType
from symphonizer.node_runner import NodeRunner

logger = logging.getLogger(__name__)


class DAGNote:
    """
    A DAGNote is a vertex or node in a DAG graph. It contains the configuration required to process
    the node as well as the processing result.

    By default, all nodes are unique with only single instance of the node in the DAG. A key attribute
    of DAGNote class is that it is hashable and can be used as a key in a dictionary or as an
    element in a set.

    There use cases where there parameters to a node make is unique, e.g. a node that is processing
    a file, the file name is a parameter to the node. In this case, on initialization, the
    single_instance parameter can be set to false. The node is then identified by the node name and
    a unique instance id.
    """
    _single_instance: bool
    _node_name: str
    _instance_id: str

    node_data: Dict[str, Any]
    error: ErrorType | None
    result: Any | None
    start_time: float | None
    end_time: float | None

    def __init__(self, node_name: str, single_instance: Optional[bool] = None, **node_data: Any):
        self._single_instance = True if single_instance is None else single_instance
        self._node_name = node_name
        self._instance_id = uuid.uuid4().hex
        self._node_data = node_data

        self.error = None
        self.result = None
        self.start_time = None
        self.end_time = None

    @property
    def single_instance(self) -> bool:
        """ Post initialization, the single_instance property is immutable.
        """
        return self._single_instance

    @property
    def instance_id(self) -> str:
        """ Post initialization, the instance_id property is immutable.
        """
        return self._instance_id

    @property
    def node_name(self) -> str:
        """ Post initialization, the node_name property is immutable.
        """
        return self._node_name

    def __str__(self) -> str:
        if self.single_instance:
            return f"{self._node_name}"
        else:
            return f"{self._node_name}:{self._instance_id}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, DAGNote):
            return self.__hash__() == other.__hash__()
        else:
            return False

    def __hash__(self) -> int:
        if self.single_instance:
            return hash(self._node_name)
        else:
            return hash(self._instance_id)


class Composition:
    """
    Composition represents a Directed Acyclic Graph (DAG) with the capabilities to iterates over
    the graph in topological order while processing each node.

    The handling failures and retry policies for nodes are outside the scope of this class, these
    important aspects are handled either through the NodeRunner implementation or broader
    implementation in which the Composition resides.

    Node scheduling considerations are:
    - Node processing retry policies?
    - Is the overall processing be stopped on node errors?
    - DAG or vertex processing timeouts policies?
    """
    _instance_id: str
    _graph: GraphType
    _schedule_done_cb: CompositionDoneFunction | None
    _node_processing_done_cb: NodeProcessDoneFunction | None
    _ts: TopologicalSorter[DAGNote]
    _task_queue: Queue[DAGNote]
    _running_tasks: Dict[DAGNote, asyncio.Future[NodeRunnerType]]
    _start_time: float
    _end_time: float | None
    _event_loop: asyncio.AbstractEventLoop

    started: bool
    running: bool
    paused: bool
    stopped: bool
    errored: BaseException | None

    def __init__(
            self,
            graph: GraphType,
            schedule_done_cb: Optional[CompositionDoneFunction] = None,
            node_processing_done_cb: Optional[NodeProcessDoneFunction] = None,
            event_loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        self.instance_id = uuid.uuid4().hex
        self._event_loop = event_loop or asyncio.get_running_loop()
        self._graph = graph or {}
        self._schedule_done_cb = schedule_done_cb
        self._node_processing_done_cb = node_processing_done_cb
        self._ts = TopologicalSorter(self._graph)
        self._task_queue = Queue()
        self._start_time = time.time()
        """ NOTE: the state of self._start_time is reset when the scheduling is started.
        """
        self._end_time = None
        self.started: bool = False
        self.running: bool = False
        self.paused: bool = False
        self.stopped: bool = False
        self.errored = None
        self._running_tasks = {}  # Index of running tasks

    @property
    def status(self) -> CompositionStatus:
        if self.stopped:
            return "cancelled"
        elif self.paused:
            return "paused"
        elif self.running:
            return "running"
        elif self.completed:
            return "completed"
        else:
            return "new"

    @property
    def completed(self) -> bool:
        """
        READONLY property,
        :return: True if all vertexes have been processed, returns False until first started
        """
        if self.started:
            return not self._ts.is_active()
        else:
            return False

    def process_node_done(self, node: DAGNote, future: asyncio.Future[Any]) -> None:
        """ Performs the cleanup after the processing of a node has ended - is done.

        The following exceptions that are raised by the underlying node processing will
        influence the workflow of the Composition:
        - asyncio.exceptions.CancelledError: async processing was cancelled
        - asyncio.exceptions.TimeoutError: asyncio processing exceeded the timeout and was ended
        - TryAgainException(Exception): raised by the NodeRunner / task executor to indicate that
            the node processing should be retried
        - ContinueAfterErrorException(Exception): raised by the NodeRunner / task executor to
            indicate that DAG interation should continue as normal after an error
        - StopScheduleException(Exception): raised by the NodeRunner / task executor to indicate
            that the DAG processing should be stopped immediately.
        """
        node.end_time = time.time()
        self._running_tasks.pop(node, None)
        error = future.exception()
        if error is None:
            node.result = future.result()
        node.error = {
            "error": type(error).__name__,
            "description": str(error),
        }
        status: NodeDoneStatus = "error"
        if error:
            # handle processing exceptions
            # logger.debug("Node processing error, %s: %s", node, error)
            if isinstance(error, TryAgainException):
                # logger.debug("Node processing, try again, %s", node)
                self.process_node(node)
                status = "retrying"

            elif isinstance(error, ContinueAfterErrorException):
                # logger.debug("Node processing, ignore error, %s", node)
                self._ts.done(node)
                "completed_error"

            elif isinstance(error, StopScheduleException):
                # logger.debug("Node processing, stop schedule, %s", node)
                self.errored = error
                self.stop_processing()
                status = "cancelled"
            else:
                # Handle general runtime errors, do not mark the node as done, add back to running tasks
                # logger.error("Node processing general error, %s", error)
                self._running_tasks[node] = future
                status = "error"

        else:
            self._ts.done(node)
            # logger.debug("Node processing complete, %s", node)
            status = "completed"

        if self._node_processing_done_cb:
            self._node_processing_done_cb(node, status, error)

    def configure_node_runner(self, node: Hashable) -> NodeRunner:
        """ Configure_node_runner is called to configure a node runner. The method is required to the
        implemented in subclasses of this class.

        :param node:
        :return: an instance of NodeRunner
        """
        _ = self, node
        node_runner = NodeRunner(
            node_executor=MockInstrument()
        )
        return node_runner

    def process_node(self, node: DAGNote) -> None:
        """ _process_node is called to process a single node.
        It's prudent that this implementation addresses likely exceptions:
        * asyncio.exceptions.CancelledError - Future / Coroutine explicitly cancelled
        * asyncio.exceptions.TimeoutError - asyncio schedule timeout

        :param node:
        :return:
        """
        node.start_time = time.time()
        node_runner = self.configure_node_runner(node)
        result_future = self._event_loop.create_task(node_runner())
        result_future.add_done_callback(functools.partial(self.process_node_done, node))
        self._running_tasks[node] = result_future

    def stop_processing(self) -> bool:
        """ Stop processing of all vertexes, allow all current processing to complete.
        :return: True if there are still vertexes to process, else return False
        """
        if self.stopped:
            return False
        else:
            self.stopped = True
            return True

    def pause_processing(self) -> bool:
        """
        Stop processing new vertexes, allow all current processing to complete.
        :return: True if there are still vertexes to process, else return False
        """
        if not self.stopped and not self.paused and self._ts.is_active():
            self.paused = True
            return True
        else:
            return False

    async def start_processing(self) -> None:
        """
        Begin processing of all unprocessed vertexes, also called to resume after processing has
        being paused. No amendments to the graph are allowed after processing has first started
        or during a pause. Once stopped, the scheduler can not be started again.

        If any cycle is detected, graphlib.CycleError will be raised.

        :return: None
        """
        if self.stopped or self.running or self.completed:
            return

        if self.paused:
            self.paused = False

        if not self.started:
            # Only call prepare once
            self._ts.prepare()
            self.started = True
            self._start_time = time.time()

        try:
            self.running = True
            list(map(self._task_queue.put_nowait, self._ts.get_ready()))
            """ NOTE: The state of self.stopped or self.paused can be updated by another coroutine or thread
            """
            while not (self.stopped or self.paused) and self._ts.is_active():
                if not self._task_queue.empty():
                    task_node = self._task_queue.get_nowait()
                    self.process_node(task_node)

                for next_node in self._ts.get_ready():
                    await self._task_queue.put(next_node)

                """This sleep is added here to force the event loop to schedule background awaitables. This is
                typically the case where there's no background async io activity, like with unittests
                or in synchronous applications.
                """
                await asyncio.sleep(0)
            await asyncio.gather(*[f for f in self._running_tasks.values()])

        except Exception as err:
            logger.error("Scheduler processing error: %s", err)
            self.errored = err
            raise

        finally:
            self.running = False
            self._end_time = time.time()
            time_elapsed = self._end_time - self._start_time
            if (not self._ts.is_active() or self.stopped) and self._schedule_done_cb:
                self._schedule_done_cb(self, self.status, self.errored, time_elapsed)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}:{self._instance_id}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, DAGNote):
            return self.__hash__() == other.__hash__()
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._instance_id)



import asyncio
from asyncio import Queue, QueueEmpty
from typing import Dict, Any
import logging
import functools

from symphonizer.composition import Composition

logger = logging.getLogger(__name__)
verbose = False


class Perform:
    """ Process multiple dag compositions concurrently
    """
    _high_water_count: int
    _low_water_count: int
    _high_water_reached: bool
    _incoming_queue: Queue[Composition]
    _running_dags: Dict[str, Composition]

    def __init__(self) -> None:
        self._high_water_count = 100
        self._low_water_count = 50
        self._high_water_reached = False
        self._incoming_queue = Queue()
        self._running_dags = {}

    def _dag_completed(self, dag: Composition, future: asyncio.Future[Any]):
        """ Handles a completed dag:Composition and schedules the next one to start. If the high water mark
        has been reached, adds the dag to the incoming queue. If the low water mark has been reached,
        schedules the next dag in the queue to be processed.
        """
        self._running_dags.pop(dag.instance_id, None)
        if verbose:
            error = future.exception()
            logger.debug("Composition dag ended %s, error: %s", dag.instance_id, error)

        if self._high_water_reached and len(self._running_dags) < self._low_water_count:
            logging.debug("Low water mark reached %s", self._low_water_count)
            self._high_water_reached = False

        if not self._high_water_reached and not self._incoming_queue.empty():
            try:
                dag = self._incoming_queue.get_nowait()
                self._start_dag(dag)
            except QueueEmpty:
                pass

    def _start_dag(self, dag: Composition) -> None:
        # logger.debug("Composition dag starting %s", dag.instance_id)
        task = asyncio.create_task(dag.start_processing())
        task.add_done_callback(functools.partial(self._dag_completed, dag))
        self._running_dags.setdefault(dag.instance_id, dag)

    async def add(self, dag: Composition) -> None:
        if dag.instance_id not in self._running_dags and not dag.running and not dag.stopped:
            if len(self._running_dags) < self._high_water_count:
                self._start_dag(dag)
                if len(self._running_dags) == self._high_water_count and not self._high_water_reached:
                    self._high_water_reached = True
                    logging.debug("High water mark reached %s", self._high_water_count)
            else:
                await self._incoming_queue.put(dag)





import asyncio
from typing import Any

class MockInstrument:
    """ A Mock Node Executor useful for testing and simulation"""

    def __init__(self, **kwargs: Any):
        self.kwargs = kwargs
        self.delay = kwargs.get("delay", 0)

    async def __call__(self, **params: Any) -> str:
        delay = ""
        if self.delay:
            await asyncio.sleep(self.delay)
            delay = " with delay of {} seconds".format(self.delay)
        return f"Completed {self.__class__.__name__} with {len(self.kwargs)} kwargs and {len(params)} parameters{delay}"



import logging
from typing import Optional, Any, Dict
from uuid import uuid4

from nuropb.rmq_api import RMQAPI

logger = logging.getLogger(__name__)

_stateful_nuropb_clients: Dict[str, RMQAPI] = {}


class NuroPbInstrument:
    _amqp_url: str
    _rpc_response: bool

    def __init__(self, amqp_url: Optional[str] = None, rpc_response: Optional[bool] = None, stateful: Optional[bool] = None):
        self._amqp_url = amqp_url or ""
        self._rpc_response = True if rpc_response is None else rpc_response
        self._stateful = False if stateful is None else stateful

    def _init_nuro_pb_client(self, amqp_url: str) -> RMQAPI:
        if self._stateful:
            if amqp_url in _stateful_nuropb_clients:
                return _stateful_nuropb_clients[amqp_url]
            else:
                _stateful_nuropb_clients[amqp_url] = RMQAPI(amqp_url=amqp_url)
                return _stateful_nuropb_clients[amqp_url]
        else:
            return RMQAPI(amqp_url=amqp_url)

    async def __call__(self, **kwargs: Any) -> Any:
        if "amqp_url" in kwargs:
            self._amqp_url = kwargs["amqp_url"]
        if "rpc_response" in kwargs:
            self._rpc_response = kwargs["rpc_response"]

        connected = False
        try:
            nuropb_client = self._init_nuro_pb_client(amqp_url=self._amqp_url)
            if not nuropb_client.connected:
                await nuropb_client.connect()
            connected = nuropb_client.connected
        except Exception as e:
            logger.debug(f"Error connecting to NuroPb service mesh: {e}")
            nuropb_client = None

        if not connected:
            raise RuntimeError("NuroPb service mesh api failed to connect")
        else:
            try:
                service = kwargs["service"]
                method = kwargs["method"]
                params = kwargs.get("params", {})
                context = kwargs.get("context", {})
                trace_id = kwargs.get("trace_id", uuid4().hex)
                ttl = kwargs.get("ttl", 60 * 60 * 1000)  # default timeout set to 1 hour
                rpc_response = kwargs.get("rpc_response", self._rpc_response)
                return await nuropb_client.request(
                    service=service,
                    method=method,
                    params=params,
                    context=context,
                    trace_id=trace_id,
                    ttl=ttl,
                    rpc_response=rpc_response,
                )
            finally:
                if not self._stateful and connected:
                    try:
                        await nuropb_client.disconnect()
                    except Exception as e:
                        _ = e
